/****************************************************************************
 * This file is part of the project AqDatabase.
 * AqDatabase (c) by 2016 Martin Preuss, all rights reserved.
 *
 * The license for this file can be found in the file COPYING which you
 * should have received along with this file.
 ****************************************************************************/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif


#include "createdb.h"
#include "checkobjects.h"
#include "createobjects.h"
#include "type1.h"
#include "type2.h"



#define RUN1_TYPE1_NUM     1000
#define RUN1_TYPE1_STARTID 10000

#define RUN1_TYPE2_NUM     800
#define RUN1_TYPE2_STARTID 20000




int test_create_objects(const char *url, const char *dirname, const char *fname) {
  AQDB_DB *db=NULL;
  AQDB_ID tableId1;
  AQDB_ID tableId2;
  int rv;

  /* create database */
  rv=create_database(url, dirname, fname);
  if (rv<0) {
    fprintf(stderr, "ERROR: Could not create database (%d)\n", rv);
    return rv;
  }


  rv=AQDB_DbFactory(url, &db);
  if (rv<0) {
    fprintf(stderr, "ERROR: Unable to instantiate db (%d)\n", rv);
    return rv;
  }

  /* open database */
  rv=AQDB_DB_Open(db, AQDB_ACTION_FLAGS_READ | AQDB_ACTION_FLAGS_WRITE);
  if (rv<0) {
    fprintf(stderr, "ERROR: Unable to open db (%d)\n", rv);
    AQDB_DB_free(db);
    return rv;
  }

  /* open table 1 */
  tableId1=0;
  rv=AQDB_DB_OpenTable(db, "type1",
                       AQDB_ACTION_FLAGS_READ | AQDB_ACTION_FLAGS_WRITE,
                       &tableId1);
  if (rv<0) {
    fprintf(stderr, "ERROR: Unable to open table for type1 (%d)\n", rv);
    AQDB_DB_free(db);
    return rv;
  }

  /* open table 2 */
  tableId2=0;
  rv=AQDB_DB_OpenTable(db, "type2",
                       AQDB_ACTION_FLAGS_READ | AQDB_ACTION_FLAGS_WRITE,
                       &tableId2);
  if (rv<0) {
    fprintf(stderr, "ERROR: Unable to open table for type2 (%d)\n", rv);
    AQDB_DB_free(db);
    return rv;
  }

  /* begin edit session */
  rv=AQDB_DB_BeginEdit(db, 0);
  if (rv<0) {
    fprintf(stderr, "ERROR: Unable to begin edit session (%d)\n", rv);
    AQDB_DB_free(db);
    return rv;
  }

  /* create type 1 objects */
  rv=create_type1_objects(db, tableId1, RUN1_TYPE1_NUM, RUN1_TYPE1_STARTID); /* create x objects, starting with 2000 */
  if (rv<0) {
    fprintf(stderr, "ERROR: Unable to create type1 objects (%d)\n", rv);
    AQDB_DB_free(db);
    return rv;
  }

  /* create type 2 objects */
  rv=create_type2_objects(db, tableId2, RUN1_TYPE2_NUM, RUN1_TYPE2_STARTID); /* create x objects, starting with 7000 */
  if (rv<0) {
    fprintf(stderr, "ERROR: Unable to create type2 objects (%d)\n", rv);
    AQDB_DB_free(db);
    return rv;
  }

  /* done editing */
  rv=AQDB_DB_EndEdit(db, 0);
  if (rv<0) {
    fprintf(stderr, "ERROR: Unable to end edit session (%d)\n", rv);
    AQDB_DB_free(db);
    return rv;
  }

  if (1) {
    TYPE1_LIST *tl;
    TYPE1 *t;
    AQDB_ID currentId;
    int num;

    tl=Type1_List_new();
    rv=read_type1_objects(db, tl, tableId1, NULL);
    if (rv<0) {
      fprintf(stderr, "ERROR: Unable to read type1 objects (%d)\n", rv);
      AQDB_DB_free(db);
      return rv;
    }

    /* sort list */
    Type1_List_SortById(tl, 1);

    t=Type1_List_First(tl);
    while(t) {
      rv=check_type1_object(db, t);
      if (rv<0) {
        fprintf(stderr, "ERROR: Bad result checking type1 object (%d)\n", rv);
        AQDB_DB_free(db);
        return rv;
      }

      t=Type1_List_Next(t);
    }

    currentId=RUN1_TYPE1_STARTID;
    num=RUN1_TYPE1_NUM;
    t=Type1_List_First(tl);
    while(t && num) {
      AQDB_ID id;

      id=Type1_GetId(t);
      if (id!=currentId) {
        fprintf(stderr, "ERROR: Unexpected type1 id (got %llu, expected %llu)\n",
                (unsigned long long) id,
                (unsigned long long) currentId);
        return GWEN_ERROR_INTERNAL;
      }
      currentId++;
      num--;
      t=Type1_List_Next(t);
    }
    if (t) {
      fprintf(stderr, "ERROR: There are more objects of type1 than expected\n");
      return GWEN_ERROR_INTERNAL;
    }
    if (num) {
      fprintf(stderr, "ERROR: There are too few objects of type1\n");
      return GWEN_ERROR_INTERNAL;
    }
  }

  if (1) {
    TYPE2_LIST *tl;
    TYPE2 *t;
    AQDB_ID currentId;
    int num;

    tl=Type2_List_new();
    rv=read_type2_objects(db, tl, tableId2, NULL);
    if (rv<0) {
      fprintf(stderr, "ERROR: Unable to read type2 objects (%d)\n", rv);
      AQDB_DB_free(db);
      return rv;
    }

    /* sort list */
    Type2_List_SortById(tl, 1);

    t=Type2_List_First(tl);
    while(t) {
      rv=check_type2_object(db, t);
      if (rv<0) {
        fprintf(stderr, "ERROR: Bad result checking type2 object (%d)\n", rv);
        AQDB_DB_free(db);
        return rv;
      }

      t=Type2_List_Next(t);
    }

    currentId=RUN1_TYPE2_STARTID;
    num=RUN1_TYPE2_NUM;
    t=Type2_List_First(tl);
    while(t && num) {
      AQDB_ID id;

      id=Type2_GetId(t);
      if (id!=currentId) {
        fprintf(stderr, "ERROR: Unexpected type2 id (got %llu, expected %llu)\n",
                (unsigned long long) id,
                (unsigned long long) currentId);
        return GWEN_ERROR_INTERNAL;
      }
      currentId++;
      num--;
      t=Type2_List_Next(t);
    }
    if (t) {
      fprintf(stderr, "ERROR: There are more objects of type2 than expected\n");
      return GWEN_ERROR_INTERNAL;
    }
    if (num) {
      fprintf(stderr, "ERROR: There are too few objects of type2\n");
      return GWEN_ERROR_INTERNAL;
    }
  }



  rv=AQDB_DB_Close(db, 0);
  if (rv<0) {
    fprintf(stderr, "ERROR: Unable to close db (%d)\n", rv);
    AQDB_DB_free(db);
    return 2;
  }

  AQDB_DB_free(db);

  fprintf(stderr, "Ok.\n");
  return 0;
}
